@@ -7,6 +7,27 @@ import json
 from django.utils.html import mark_safe
 from django_pandas.io import read_frame
 
+
+class FixedTextColumn(tables.Column):
+    ''' Mimics behavior of django_tables2.tables.Column but allows for a fixed text to be rendered '''
+    def __init__(self, *args, text='', empty_values=None, **kwargs):
+        if empty_values == None:
+            empty_values = []
+        super().__init__(*args, empty_values=empty_values, **kwargs)
+        self.text = text
+
+    def text_value(self, record, value):
+        if self.text is None:
+            return value
+        return self.text(record) if callable(self.text) else self.text
+
+    def value(self, record, value):
+        return self.text_value(record, value)
+
+    def render(self, record, value):
+        return self.text_value(record, value)
+
+
 class CheckFkColumn(tables.Column):
     ''' A column checks and displays the existence of a FK relationship '''
     def __init__(self, *args, fk_attr=None, present_symbol='✔', absent_symbol='X', **kwargs):
@@ -21,13 +42,24 @@ class CheckFkColumn(tables.Column):
         return f'{value}{check}'
 
 class RoundNumberColumn(tables.Column):
-    ''' A column that will round a number and add commas for display '''
+    """ A column that will round a number and add commas for display
 
-    def __init__(self, *args, money=False, round_to=2, prefix='', **kwargs):
+    Args:
+        *args: arguments to be passed to django-tables2 Column class __init__ method
+        money: (bool default=False) Will append a dollar sign $ to front of value if true
+        round_to: (int default=2) The number of decimals to round to
+        prefix: (str, optional) This string is concatenated to the front of value
+        suffix: (str, optional) This string is concatenated to the end of value
+        orderable: (bool default=True) Can django-table sort/order this col. Will be set to false if money, suffix or prefix provided
+        **kwargs: arguments to be passed to django-tables2 Column class __init__ method
+    """    
+
+    def __init__(self, *args, money=False, round_to=2, prefix='', suffix='', **kwargs):
         super().__init__(*args, **kwargs)
         self.money = money
         self.round_to = round_to
         self.prefix = prefix
+        self.suffix = suffix
     
     def render(self, value, record):
         val = round(value, self.round_to)
@@ -37,19 +69,22 @@ class RoundNumberColumn(tables.Column):
             rstr = f'{val:,.{self.round_to}f}'
         if self.money: 
             rstr = '$' + rstr
-        return self.prefix + rstr
+        return self.prefix + rstr + self.suffix
 
+    def value(self, value, record):
+        ''' Overrides default value method (that would just call render on table export) '''
+        return value
       
 class CollapseColumnBase(tables.Column):
     """ Base class for CollapseColumn, extends django-tables2 Column class
     Args:
         *args: arguments to be passed to djang-tables2 derived Column __init__ method
-        **kwargs: arguments to be passed to djang-tables2 derived Column __init__ method
-        label (:obj:`str`, optional): The string to be displayed on the collapse target. Defaults to "show"
-        label_accessor (:obj:`str`, optional): Passed to django-tables2 Accessor object along with the cell value. 
+        label (str, optional): The string to be displayed on the collapse target. Defaults to "show"
+        label_accessor (str, optional): Passed to django-tables2 Accessor object along with the cell value. 
             Result used as collapse target label
-        label_extra (:obj:`str`, optional): Extra text that will be appended to the collapse target label
-        style (:obj:`str`, optional): css expression that will be passed to the collapasable divs style parameter
+        label_extra (str, optional): Extra text that will be appended to the collapse target label
+        style (str, optional): css expression that will be passed to the collapasable divs style parameter
+        **kwargs: arguments to be passed to djang-tables2 derived Column __init__ method
     """    
     def __init__(
         self, 
@@ -75,6 +110,16 @@ class CollapseColumnBase(tables.Column):
         return f'"{r}"'
 
     def get_label(self, value=None, record=None, val=None):
+        """ A method that returns the label to be used for the collapasble div
+
+        Args:
+            value (str, optional): The value of the cell in the table (standard arg from django-tables2). Defaults to None.
+            record (_type_, optional): The record representing the row of the table (standard arg from django-tables2). Defaults to None.
+            val (_type_, optional): the html val to be rendered in the collapasable div. Defaults to None.
+
+        Returns:
+            str: The collapsable div label
+        """        
         if self.label_accessor:
             rval = A(self.label_accessor).resolve(record)
         else:
@@ -87,6 +132,17 @@ class CollapseColumnBase(tables.Column):
         return str(rval) + self.label_extra
 
     def final_render(self, value, record, val):
+        """ The final html to be rendered in the cell of the table
+
+        Args:
+            value (str, optional): The value of the cell in the table (standard arg from django-tables2). Defaults to None.
+            record (_type_, optional): The record representing the row of the table (standard arg from django-tables2). Defaults to None.
+            val (_type_, optional): the html val to be rendered in the collapasable div. Defaults to None.
+
+        Returns:
+            str: the html to be rendered in the cell of the table
+        """        
+
         randnum = random.randint(1, 1_000_000_000)
         label = self.get_label(value=value, record=record, val=val)
         if label != '':
@@ -108,11 +164,20 @@ class CollapseDictColumn(CollapseColumnBase):
     """Custom django-tables2 column that will render a dictionary in a collapsable div.
 
     Args:
-        *args: passed to CollapseColumnBase
-        sort_by: None, 'key' or 'value' determines what to sort the dictionary by, defaults to None. keys and values will be pandas astype('string')
-        ascending: bool that sorts keys/values ascending (True) or descending (False). Default = True
-        na_position: 'first' or 'last' places records with null values up front or at end respectively
-        **kwargs: passed to CollapseColumnBase
+        *args (iterable, optional): arguments to be passed to django_tables2 Column (See help(django_tables2.Column) for options)
+        *kwargs (iterable, optional): keyword arguments to be passed to django_tables2 Column (See help(django_tables2.Column) for options)
+        label (str, optional): The string to be displayed on the collapse target. Defaults to "show"
+        label_accessor (str, optional): Passed to django-tables2 Accessor object along with the cell value. 
+            Result used as collapse target label
+        label_extra (str, optional): Extra text that will be appended to the collapse target label
+        style (str, optional): css expression that will be passed to the collapasable divs style parameter
+        sort_by (str, optional): None, 'key' or 'value' determines what to sort the dictionary by. 
+            keys and values will be pandas astype('string'). defaults to None. 
+        ascending (bool, optional): bool that sorts keys/values ascending (True) or descending (False). Default = True
+        na_position (str, optional): 'first' or 'last' places records with null values up front or at end respectively
+        to_html_kwargs (dict, optional): kwargs to be passed to pd.DataFrame.to_html method. 
+            defaults to dict(classes = ['table-bordered', 'table-striped', 'table-sm'], index=False, justify='left')   
+        to_html_kwargs_extra (dict, optional): kwargs to be added to the to_html_kwargs (typically used for "adding" to defaults). Defaults to {}      
     """    
       
     def __init__(
@@ -140,10 +205,13 @@ class CollapseDictColumn(CollapseColumnBase):
         self.to_html_kwargs.update(self.to_html_kwargs_extra)
         
     def render(self, value, record):
-        val = self.get_dictionary_val(value=value)        
+        val = self.get_dictionary_html(value=value)        
         return self.final_render(value=value, record=record, val=val)
 
-    def get_dictionary_val(self, value):
+    def value(self, value, record):
+        return value
+
+    def get_dictionary_html(self, value):
         if isna(value):
             value = {}        
         if type(value) != dict:
@@ -254,13 +322,18 @@ class CollapseDataFrameColumn(CollapseColumnBase):
         qs = qs.order_by(*self.order_by_args)
         return qs if self.limit==None else qs[:self.limit]
 
-    def get_df_html(self, qs):
+    def get_df_final(self, qs):
+        ''' Final steps to pd.DF before render or export '''
         if self.use_read_frame:
             df = read_frame(qs, **self.get_read_frame_kwargs())
         else:
             df = DF(qs)
             if self.column_names:
                 df.columns = self.column_names
+        return df
+
+    def get_df_html(self, qs):
+        df = self.get_df_final(qs)
         return df.to_html(**self.to_html_kwargs)
 
     def render(self, value, record):
@@ -271,6 +344,119 @@ class CollapseDataFrameColumn(CollapseColumnBase):
             val= self.get_df_html(qs)
         return self.final_render(value=value, record=record, val=val)
 
+    def value(self, value, record):
+        ''' Return the value used during table export '''
+        qs = self.get_queryset(value)
+        if qs.count() == 0:
+            return None
+        return self.get_df_final(qs).to_dict()
+
+class CollapseIterableColumn(CollapseColumnBase):
+    """ Custom django-tables2 column that will render an iterable in a collapsable div.
+
+    Args:
+        *args (iterable, optional): arguments to be passed to django_tables2 Column (See help(django_tables2.Column) for options)
+        order_by: (*fields, default None) Expression fed to sort_by django method. Should be comma separated fields or other order_by expressions
+        hyperlink: (bool, default False) If true, will attempt to linkify the elements of the iterable
+        fkwargs: (**kwargs, default None) Lookup/field parameters used in a Django Quereyset filter
+        href_attr: (str, default None) Should be the name of an attribute or field that contains the url for linkified values
+        str_attr: (str, default None) Name of attribute that will be used for display in lieu of value's __str__()
+        **kwargs (iterable, optional): keyword arguments to be passed to django_tables2 Column (See help(django_tables2.Column) for options)
+    """    
+
+    def __init__(self, 
+        *args, 
+        order_by = None,
+        hyperlink = False,
+        fkwargs = None,
+        href_attr = None, 
+        str_attr = None,  
+        **kwargs
+    ):
+        super().__init__(*args, **kwargs)
+        self.hyperlink = hyperlink
+        self.href_attr = href_attr
+        self.str_attr = str_attr
+        self.order_by = order_by
+        self.fkwargs = fkwargs
+
+    def get_href(self, obj):
+        ''' Method derives the href value to be used in hyperlinking list items '''
+        if self.href_attr == None:
+            return obj.get_absolute_url()
+        else:
+            return getattr(obj, self.href_attr)
+
+    def get_prepped_value(self, value):
+        if self.order_by:
+            value = value.order_by(self.order_by)
+        if self.fkwargs:
+            value = value.filter(**self.fkwargs)
+        return value
+
+    def get_final_value(self, value):
+        value = self.get_prepped_value(value)
+        val = ''
+        style = self.get_style()
+        for obj in value:
+            obj_val = str(obj) if self.str_attr == None else getattr(obj, self.str_attr)
+            if self.hyperlink:
+                href = self.get_href(obj)
+                obj_val = f'<a href={href}>{obj_val}</a>'
+            val = val + f'<li style={style}>{obj_val}</li>'
+        return val
+
+    def render(self, value, record):
+        val = self.get_final_value(value)
+        return self.final_render(value=value, record=record, val=val)
+
+    def value(self, value, record):
+        val = self.get_prepped_value(value)
+        if self.str_attr:
+            alt_values = []
+            for obj in val:
+                obj_val = getattr(obj, self.str_attr)
+                alt_values.append(obj_val)
+            return alt_values
+        return list(val)
+
+class CollapseNoniterableColumn(CollapseColumnBase):
+    """ Custom django-tables2 column that will render an an object in a collapsable div.
+
+    Args:
+        *args (iterable, optional): arguments to be passed to django_tables2 Column (See help(django_tables2.Column) for options)
+        hyperlink: (bool, default False) If true, will attempt to linkify the elements of the iterable
+        href_attr: (str, default None) Should be the name of an attribute or field that contains the url for linkified values
+        **kwargs (iterable, optional): keyword arguments to be passed to django_tables2 Column (See help(django_tables2.Column) for options)
+    """   
+
+    def __init__(self, *args, hyperlink=False, href_attr=None, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.hyperlink = hyperlink
+        self.href_attr = href_attr
+
+    def get_href(self, obj):
+        ''' Method derives the href value to be used in hyperlinking list items '''
+        if self.href_attr == None:
+            return obj.get_absolute_url()
+        else:
+            return getattr(obj, self.href_attr)
+
+    def get_prepped_value(self, value, record):
+        if self.hyperlink:
+            href = self.get_href(record)
+            val = f'<a href={href}>{value}</a>'
+        else:
+            val = value
+        return f'<div style={self.get_style()}>{val}</div>'
+
+    def render(self, value, record):
+        val = self.get_prepped_value(value=value, record=record)
+        return self.final_render(value=value, record=record, val=val)   
+
+    def value(self, value, record):
+        return value
+
 class CollapseColumn(CollapseColumnBase):
     ''' Column is meant for columns that have lots of data in each cell to make viewing cleaner'''
 
